{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Maze Solver with A* Algorithm\n",
    "\n",
    "![Maze Solver A* Source: Wikipedia](resources/011-maze_solved.png)\n",
    "\n",
    "[A* Algorithm](https://en.wikipedia.org/wiki/A*_search_algorithm) give always an optimal solution based on a [bread-first-search](https://en.wikipedia.org/wiki/Breadth-first_search) and queues. It is an extension over the previous [Lee's Algorithms](009-maze-solver-lee.ipynb) and [Dijkstra's Algorithms](010-maze-solver-dijkstra.ipynb), and uses heuristics to guide its search.\n",
    "\n",
    "A* Algorithm has the in the worst case the same time complexity as Dijkstra which is depending on the queue implementation, where $|E|$ = number of edges and $|V|$ number of nodes.\n",
    "\n",
    "* Fibonacci heap: $O(|E| + |V| log |V|)$\n",
    "* binary heap: $O((|E| + |V|) log |V|)$\n",
    "* unsorted array: $O(|V|^2)$\n",
    "\n",
    "But A* is based on heuristics which helps to find a more optimal solution much quicker.\n",
    "\n",
    "![A* Algorithms Demo Source: Wikipedia](resources/011-astar_maze_animation.gif)\n",
    "\n",
    "## Terminology\n",
    "\n",
    "* Node or State - All uniquely identifiable elements where it's possible to stop\n",
    "* Transition - Moving between states\n",
    "* Source Node - Location to start the search\n",
    "* Destination Node - Location to be found\n",
    "* Search Space - A collection of nodes\n",
    "* Cost - Numerical value representing the cost for moving vom one node to another\n",
    "\n",
    "### Node Values\n",
    "\n",
    "One important aspect of A* is the cost calculations. $f(n) = g(n)+h(n)$. Every time a new node is created these values are calculated:\n",
    "* $f(n)$ = Total cost of a node\n",
    "* $g(n)$ = distance between current node and start node\n",
    "* $h(n)$ = heuristic, estimated distance between current node and the end node\n",
    "\n",
    "## Fibonacci Heap\n",
    "\n",
    "With Heap Collection method such as Binomial or Fibonacci Heap to collect the trees the algorithm can be more time efficient.\n",
    "\n",
    "[Fibonacci Head](https://en.wikipedia.org/wiki/Fibonacci_heap) is a datastructure for priority queue operation. \n",
    "Fibonacci Heap maintains a pointer to minimum value (which is root of a tree). All tree roots are connected using circular doubly linked list, so all of them can be accessed using single ‘min’ pointer. \n",
    "\n",
    "![Fibonacci Heap Source: Wikipedia](resources/010-fibonacci_heap.png)\n",
    "\n",
    "## Solution\n",
    "\n",
    "Flowchart of the solution\n",
    "\n",
    "![A* Flowchart Source: Wikipedia](resources/011-astar_flowchart.png)\n",
    "\n",
    "### Pseudo code\n",
    "\n",
    "```\n",
    "// A* (star) Pathfinding\n",
    "// Initialize both open and closed list\n",
    "let the openList equal empty list of nodes\n",
    "let the closedList equal empty list of nodes\n",
    "// Add the start node\n",
    "put the startNode on the openList (leave it's f at zero)\n",
    "// Loop until you find the end\n",
    "while the openList is not empty\n",
    "    // Get the current node\n",
    "    let the currentNode equal the node with the least f value\n",
    "    remove the currentNode from the openList\n",
    "    add the currentNode to the closedList\n",
    "    // Found the goal\n",
    "    if currentNode is the goal\n",
    "        Congrat! You've found the end! Backtrack to get path\n",
    "    // Generate children\n",
    "    let the children of the currentNode equal the adjacent nodes\n",
    "    \n",
    "    for each child in the children\n",
    "        // Child is on the closedList\n",
    "        if child is in the closedList\n",
    "            continue to beginning of for loop\n",
    "        // Create the f, g, and h values\n",
    "        child.g = currentNode.g + distance between child and current\n",
    "        child.h = distance from child to end\n",
    "        child.f = child.g + child.h\n",
    "        // Child is already in openList\n",
    "        if child.position is in the openList's nodes positions\n",
    "            if the child.g is higher than the openList node's g\n",
    "                continue to beginning of for loop\n",
    "        // Add the child to the openList\n",
    "        add the child to the openList\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Imports"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 105,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Generated array maze=\n",
      "[[0 0 0 ... 0 0 0]\n",
      " [0 0 1 ... 0 0 0]\n",
      " [0 1 1 ... 1 0 0]\n",
      " ...\n",
      " [0 1 1 ... 1 0 0]\n",
      " [0 0 1 ... 1 0 0]\n",
      " [0 0 0 ... 0 0 0]]\n"
     ]
    },
    {
     "data": {
      "text/html": [
       "\n",
       "    <div class=\"bk-root\">\n",
       "        <a href=\"https://bokeh.org\" target=\"_blank\" class=\"bk-logo bk-logo-small bk-logo-notebook\"></a>\n",
       "        <span id=\"8882\">Loading BokehJS ...</span>\n",
       "    </div>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "application/javascript": "\n(function(root) {\n  function now() {\n    return new Date();\n  }\n\n  const force = true;\n\n  if (typeof root._bokeh_onload_callbacks === \"undefined\" || force === true) {\n    root._bokeh_onload_callbacks = [];\n    root._bokeh_is_loading = undefined;\n  }\n\n  const JS_MIME_TYPE = 'application/javascript';\n  const HTML_MIME_TYPE = 'text/html';\n  const EXEC_MIME_TYPE = 'application/vnd.bokehjs_exec.v0+json';\n  const CLASS_NAME = 'output_bokeh rendered_html';\n\n  /**\n   * Render data to the DOM node\n   */\n  function render(props, node) {\n    const script = document.createElement(\"script\");\n    node.appendChild(script);\n  }\n\n  /**\n   * Handle when an output is cleared or removed\n   */\n  function handleClearOutput(event, handle) {\n    const cell = handle.cell;\n\n    const id = cell.output_area._bokeh_element_id;\n    const server_id = cell.output_area._bokeh_server_id;\n    // Clean up Bokeh references\n    if (id != null && id in Bokeh.index) {\n      Bokeh.index[id].model.document.clear();\n      delete Bokeh.index[id];\n    }\n\n    if (server_id !== undefined) {\n      // Clean up Bokeh references\n      const cmd_clean = \"from bokeh.io.state import curstate; print(curstate().uuid_to_server['\" + server_id + \"'].get_sessions()[0].document.roots[0]._id)\";\n      cell.notebook.kernel.execute(cmd_clean, {\n        iopub: {\n          output: function(msg) {\n            const id = msg.content.text.trim();\n            if (id in Bokeh.index) {\n              Bokeh.index[id].model.document.clear();\n              delete Bokeh.index[id];\n            }\n          }\n        }\n      });\n      // Destroy server and session\n      const cmd_destroy = \"import bokeh.io.notebook as ion; ion.destroy_server('\" + server_id + \"')\";\n      cell.notebook.kernel.execute(cmd_destroy);\n    }\n  }\n\n  /**\n   * Handle when a new output is added\n   */\n  function handleAddOutput(event, handle) {\n    const output_area = handle.output_area;\n    const output = handle.output;\n\n    // limit handleAddOutput to display_data with EXEC_MIME_TYPE content only\n    if ((output.output_type != \"display_data\") || (!Object.prototype.hasOwnProperty.call(output.data, EXEC_MIME_TYPE))) {\n      return\n    }\n\n    const toinsert = output_area.element.find(\".\" + CLASS_NAME.split(' ')[0]);\n\n    if (output.metadata[EXEC_MIME_TYPE][\"id\"] !== undefined) {\n      toinsert[toinsert.length - 1].firstChild.textContent = output.data[JS_MIME_TYPE];\n      // store reference to embed id on output_area\n      output_area._bokeh_element_id = output.metadata[EXEC_MIME_TYPE][\"id\"];\n    }\n    if (output.metadata[EXEC_MIME_TYPE][\"server_id\"] !== undefined) {\n      const bk_div = document.createElement(\"div\");\n      bk_div.innerHTML = output.data[HTML_MIME_TYPE];\n      const script_attrs = bk_div.children[0].attributes;\n      for (let i = 0; i < script_attrs.length; i++) {\n        toinsert[toinsert.length - 1].firstChild.setAttribute(script_attrs[i].name, script_attrs[i].value);\n        toinsert[toinsert.length - 1].firstChild.textContent = bk_div.children[0].textContent\n      }\n      // store reference to server id on output_area\n      output_area._bokeh_server_id = output.metadata[EXEC_MIME_TYPE][\"server_id\"];\n    }\n  }\n\n  function register_renderer(events, OutputArea) {\n\n    function append_mime(data, metadata, element) {\n      // create a DOM node to render to\n      const toinsert = this.create_output_subarea(\n        metadata,\n        CLASS_NAME,\n        EXEC_MIME_TYPE\n      );\n      this.keyboard_manager.register_events(toinsert);\n      // Render to node\n      const props = {data: data, metadata: metadata[EXEC_MIME_TYPE]};\n      render(props, toinsert[toinsert.length - 1]);\n      element.append(toinsert);\n      return toinsert\n    }\n\n    /* Handle when an output is cleared or removed */\n    events.on('clear_output.CodeCell', handleClearOutput);\n    events.on('delete.Cell', handleClearOutput);\n\n    /* Handle when a new output is added */\n    events.on('output_added.OutputArea', handleAddOutput);\n\n    /**\n     * Register the mime type and append_mime function with output_area\n     */\n    OutputArea.prototype.register_mime_type(EXEC_MIME_TYPE, append_mime, {\n      /* Is output safe? */\n      safe: true,\n      /* Index of renderer in `output_area.display_order` */\n      index: 0\n    });\n  }\n\n  // register the mime type if in Jupyter Notebook environment and previously unregistered\n  if (root.Jupyter !== undefined) {\n    const events = require('base/js/events');\n    const OutputArea = require('notebook/js/outputarea').OutputArea;\n\n    if (OutputArea.prototype.mime_types().indexOf(EXEC_MIME_TYPE) == -1) {\n      register_renderer(events, OutputArea);\n    }\n  }\n\n  \n  if (typeof (root._bokeh_timeout) === \"undefined\" || force === true) {\n    root._bokeh_timeout = Date.now() + 5000;\n    root._bokeh_failed_load = false;\n  }\n\n  const NB_LOAD_WARNING = {'data': {'text/html':\n     \"<div style='background-color: #fdd'>\\n\"+\n     \"<p>\\n\"+\n     \"BokehJS does not appear to have successfully loaded. If loading BokehJS from CDN, this \\n\"+\n     \"may be due to a slow or bad network connection. Possible fixes:\\n\"+\n     \"</p>\\n\"+\n     \"<ul>\\n\"+\n     \"<li>re-rerun `output_notebook()` to attempt to load from CDN again, or</li>\\n\"+\n     \"<li>use INLINE resources instead, as so:</li>\\n\"+\n     \"</ul>\\n\"+\n     \"<code>\\n\"+\n     \"from bokeh.resources import INLINE\\n\"+\n     \"output_notebook(resources=INLINE)\\n\"+\n     \"</code>\\n\"+\n     \"</div>\"}};\n\n  function display_loaded() {\n    const el = document.getElementById(\"8882\");\n    if (el != null) {\n      el.textContent = \"BokehJS is loading...\";\n    }\n    if (root.Bokeh !== undefined) {\n      if (el != null) {\n        el.textContent = \"BokehJS \" + root.Bokeh.version + \" successfully loaded.\";\n      }\n    } else if (Date.now() < root._bokeh_timeout) {\n      setTimeout(display_loaded, 100)\n    }\n  }\n\n\n  function run_callbacks() {\n    try {\n      root._bokeh_onload_callbacks.forEach(function(callback) {\n        if (callback != null)\n          callback();\n      });\n    } finally {\n      delete root._bokeh_onload_callbacks\n    }\n    console.debug(\"Bokeh: all callbacks have finished\");\n  }\n\n  function load_libs(css_urls, js_urls, callback) {\n    if (css_urls == null) css_urls = [];\n    if (js_urls == null) js_urls = [];\n\n    root._bokeh_onload_callbacks.push(callback);\n    if (root._bokeh_is_loading > 0) {\n      console.debug(\"Bokeh: BokehJS is being loaded, scheduling callback at\", now());\n      return null;\n    }\n    if (js_urls == null || js_urls.length === 0) {\n      run_callbacks();\n      return null;\n    }\n    console.debug(\"Bokeh: BokehJS not loaded, scheduling load and callback at\", now());\n    root._bokeh_is_loading = css_urls.length + js_urls.length;\n\n    function on_load() {\n      root._bokeh_is_loading--;\n      if (root._bokeh_is_loading === 0) {\n        console.debug(\"Bokeh: all BokehJS libraries/stylesheets loaded\");\n        run_callbacks()\n      }\n    }\n\n    function on_error(url) {\n      console.error(\"failed to load \" + url);\n    }\n\n    for (let i = 0; i < css_urls.length; i++) {\n      const url = css_urls[i];\n      const element = document.createElement(\"link\");\n      element.onload = on_load;\n      element.onerror = on_error.bind(null, url);\n      element.rel = \"stylesheet\";\n      element.type = \"text/css\";\n      element.href = url;\n      console.debug(\"Bokeh: injecting link tag for BokehJS stylesheet: \", url);\n      document.body.appendChild(element);\n    }\n\n    for (let i = 0; i < js_urls.length; i++) {\n      const url = js_urls[i];\n      const element = document.createElement('script');\n      element.onload = on_load;\n      element.onerror = on_error.bind(null, url);\n      element.async = false;\n      element.src = url;\n      console.debug(\"Bokeh: injecting script tag for BokehJS library: \", url);\n      document.head.appendChild(element);\n    }\n  };\n\n  function inject_raw_css(css) {\n    const element = document.createElement(\"style\");\n    element.appendChild(document.createTextNode(css));\n    document.body.appendChild(element);\n  }\n\n  \n  const js_urls = [\"https://cdn.bokeh.org/bokeh/release/bokeh-2.4.2.min.js\", \"https://cdn.bokeh.org/bokeh/release/bokeh-gl-2.4.2.min.js\", \"https://cdn.bokeh.org/bokeh/release/bokeh-widgets-2.4.2.min.js\", \"https://cdn.bokeh.org/bokeh/release/bokeh-tables-2.4.2.min.js\", \"https://cdn.bokeh.org/bokeh/release/bokeh-mathjax-2.4.2.min.js\"];\n  const css_urls = [];\n  \n\n  const inline_js = [\n    function(Bokeh) {\n      Bokeh.set_log_level(\"info\");\n    },\n    function(Bokeh) {\n    \n    \n    }\n  ];\n\n  function run_inline_js() {\n    \n    if (root.Bokeh !== undefined || force === true) {\n      \n    for (let i = 0; i < inline_js.length; i++) {\n      inline_js[i].call(root, root.Bokeh);\n    }\n    if (force === true) {\n        display_loaded();\n      }} else if (Date.now() < root._bokeh_timeout) {\n      setTimeout(run_inline_js, 100);\n    } else if (!root._bokeh_failed_load) {\n      console.log(\"Bokeh: BokehJS failed to load within specified timeout.\");\n      root._bokeh_failed_load = true;\n    } else if (force !== true) {\n      const cell = $(document.getElementById(\"8882\")).parents('.cell').data().cell;\n      cell.output_area.append_execute_result(NB_LOAD_WARNING)\n    }\n\n  }\n\n  if (root._bokeh_is_loading === 0) {\n    console.debug(\"Bokeh: BokehJS loaded, going straight to plotting\");\n    run_inline_js();\n  } else {\n    load_libs(css_urls, js_urls, function() {\n      console.debug(\"Bokeh: BokehJS plotting callback run at\", now());\n      run_inline_js();\n    });\n  }\n}(window));",
      "application/vnd.bokehjs_load.v0+json": ""
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "  <div class=\"bk-root\" id=\"000c7db5-5cc2-48b9-bcca-8fed60624189\" data-root-id=\"8883\"></div>\n"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "application/javascript": "(function(root) {\n  function embed_document(root) {\n    \n  const docs_json = {\"63797384-4cb7-49d3-95bb-7035cc245c5e\":{\"defs\":[],\"roots\":{\"references\":[{\"attributes\":{\"below\":[{\"id\":\"8892\"}],\"center\":[{\"id\":\"8895\"},{\"id\":\"8899\"}],\"height\":400,\"left\":[{\"id\":\"8896\"}],\"outline_line_color\":\"#ffffff\",\"renderers\":[{\"id\":\"8924\"}],\"title\":{\"id\":\"9245\"},\"toolbar\":{\"id\":\"8907\"},\"width\":800,\"x_range\":{\"id\":\"8884\"},\"x_scale\":{\"id\":\"8888\"},\"y_range\":{\"id\":\"8886\"},\"y_scale\":{\"id\":\"8890\"}},\"id\":\"8883\",\"subtype\":\"Figure\",\"type\":\"Plot\"},{\"attributes\":{},\"id\":\"9249\",\"type\":\"AllLabels\"},{\"attributes\":{\"data\":{\"image\":[[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],[0,0,1,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,0,0,0],[0,1,1,0,1,0,1,0,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,0,0],[0,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,0],[0,0,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,1,0],[0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,1,0,0],[0,0,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,0,1,0,1,1,1,1,1,1,0],[0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,0],[0,1,1,1,1,0,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,0,1,0,0],[0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,1,0,1,0,0],[0,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,0,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,1,0],[0,0,1,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,1,0,1,0,0],[0,1,1,1,1,0,1,0,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,1,0],[0,0,0,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,0],[0,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,0,1,1,0],[0,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,0],[0,1,1,1,1,1,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,0,1,0,1,0,1,1,1,0,1,1,1,0,1,0,1,0,1,0,1,0,1,1,1,0,1,1,1,1,1,0,1,1,0],[0,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,0,0,0],[0,1,1,0,1,1,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,0,1,0,1,0,1,0,1,1,1,1,0],[0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,0],[0,0,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,1,0],[0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,0,0,0],[0,1,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,1,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,0,0],[0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,0,0,1,0,0],[0,0,1,1,1,1,1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,0,1,0,1,1,1,0,1,0,0],[0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,1,0,1,0,0],[0,1,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,0,1,1,1,0,1,1,0],[0,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,0],[0,1,1,1,1,1,1,0,1,1,1,0,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,0,1,0,1,1,0],[0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,0],[0,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,0,1,0,1,1,1,0,1,1,1,1,1,1,1,0,1,0,1,0,1,0,1,0,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,0,0],[0,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0],[0,0,1,1,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,0,1,0,1,0,1,0,1,1,1,1,1,1,1,0,1,1,1,0,0],[0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,0],[0,1,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,1,1,0,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,0,1,0,1,1,1,1,1,0,0],[0,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0],[0,0,1,1,1,0,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,0,0],[0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,0],[0,1,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,1,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,0,1,0,0],[0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0],[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]]]},\"selected\":{\"id\":\"9254\"},\"selection_policy\":{\"id\":\"9253\"}},\"id\":\"8914\",\"type\":\"ColumnDataSource\"},{\"attributes\":{},\"id\":\"8884\",\"type\":\"Range1d\"},{\"attributes\":{\"color_mapper\":{\"id\":\"8916\"},\"dh\":{\"value\":1},\"dw\":{\"value\":1},\"image\":{\"field\":\"image\"},\"x\":{\"value\":0},\"y\":{\"value\":0}},\"id\":\"8915\",\"type\":\"Image\"},{\"attributes\":{},\"id\":\"8901\",\"type\":\"WheelZoomTool\"},{\"attributes\":{},\"id\":\"8886\",\"type\":\"Range1d\"},{\"attributes\":{\"axis\":{\"id\":\"8896\"},\"coordinates\":null,\"dimension\":1,\"group\":null,\"ticker\":null},\"id\":\"8899\",\"type\":\"Grid\"},{\"attributes\":{},\"id\":\"8893\",\"type\":\"BasicTicker\"},{\"attributes\":{\"palette\":[\"#228b22\",\"#ffffff\"]},\"id\":\"8916\",\"type\":\"LinearColorMapper\"},{\"attributes\":{},\"id\":\"8905\",\"type\":\"HelpTool\"},{\"attributes\":{},\"id\":\"9248\",\"type\":\"BasicTickFormatter\"},{\"attributes\":{\"axis\":{\"id\":\"8892\"},\"coordinates\":null,\"group\":null,\"ticker\":null},\"id\":\"8895\",\"type\":\"Grid\"},{\"attributes\":{\"coordinates\":null,\"formatter\":{\"id\":\"9248\"},\"group\":null,\"major_label_policy\":{\"id\":\"9249\"},\"ticker\":{\"id\":\"8897\"},\"visible\":false},\"id\":\"8896\",\"type\":\"LinearAxis\"},{\"attributes\":{},\"id\":\"8897\",\"type\":\"BasicTicker\"},{\"attributes\":{\"coordinates\":null,\"formatter\":{\"id\":\"9251\"},\"group\":null,\"major_label_policy\":{\"id\":\"9252\"},\"ticker\":{\"id\":\"8893\"},\"visible\":false},\"id\":\"8892\",\"type\":\"LinearAxis\"},{\"attributes\":{},\"id\":\"8888\",\"type\":\"LinearScale\"},{\"attributes\":{\"color_mapper\":{\"id\":\"8919\"},\"dh\":{\"value\":1},\"dw\":{\"value\":1},\"global_alpha\":{\"value\":0.1},\"image\":{\"field\":\"image\"},\"x\":{\"value\":0},\"y\":{\"value\":0}},\"id\":\"8918\",\"type\":\"Image\"},{\"attributes\":{\"palette\":[\"#228b22\",\"#ffffff\"]},\"id\":\"8919\",\"type\":\"LinearColorMapper\"},{\"attributes\":{\"color_mapper\":{\"id\":\"8922\"},\"dh\":{\"value\":1},\"dw\":{\"value\":1},\"global_alpha\":{\"value\":0.2},\"image\":{\"field\":\"image\"},\"x\":{\"value\":0},\"y\":{\"value\":0}},\"id\":\"8921\",\"type\":\"Image\"},{\"attributes\":{\"palette\":[\"#228b22\",\"#ffffff\"]},\"id\":\"8922\",\"type\":\"LinearColorMapper\"},{\"attributes\":{\"coordinates\":null,\"data_source\":{\"id\":\"8914\"},\"glyph\":{\"id\":\"8915\"},\"group\":null,\"hover_glyph\":null,\"muted_glyph\":{\"id\":\"8921\"},\"nonselection_glyph\":{\"id\":\"8918\"},\"view\":{\"id\":\"8925\"}},\"id\":\"8924\",\"type\":\"GlyphRenderer\"},{\"attributes\":{\"source\":{\"id\":\"8914\"}},\"id\":\"8925\",\"type\":\"CDSView\"},{\"attributes\":{},\"id\":\"8900\",\"type\":\"PanTool\"},{\"attributes\":{},\"id\":\"8903\",\"type\":\"SaveTool\"},{\"attributes\":{},\"id\":\"8890\",\"type\":\"LinearScale\"},{\"attributes\":{},\"id\":\"9253\",\"type\":\"UnionRenderers\"},{\"attributes\":{\"bottom_units\":\"screen\",\"coordinates\":null,\"fill_alpha\":0.5,\"fill_color\":\"lightgrey\",\"group\":null,\"left_units\":\"screen\",\"level\":\"overlay\",\"line_alpha\":1.0,\"line_color\":\"black\",\"line_dash\":[4,4],\"line_width\":2,\"right_units\":\"screen\",\"syncable\":false,\"top_units\":\"screen\"},\"id\":\"8906\",\"type\":\"BoxAnnotation\"},{\"attributes\":{\"coordinates\":null,\"group\":null},\"id\":\"9245\",\"type\":\"Title\"},{\"attributes\":{},\"id\":\"9252\",\"type\":\"AllLabels\"},{\"attributes\":{},\"id\":\"9254\",\"type\":\"Selection\"},{\"attributes\":{},\"id\":\"9251\",\"type\":\"BasicTickFormatter\"},{\"attributes\":{\"overlay\":{\"id\":\"8906\"}},\"id\":\"8902\",\"type\":\"BoxZoomTool\"},{\"attributes\":{},\"id\":\"8904\",\"type\":\"ResetTool\"},{\"attributes\":{\"tools\":[{\"id\":\"8900\"},{\"id\":\"8901\"},{\"id\":\"8902\"},{\"id\":\"8903\"},{\"id\":\"8904\"},{\"id\":\"8905\"}]},\"id\":\"8907\",\"type\":\"Toolbar\"}],\"root_ids\":[\"8883\"]},\"title\":\"Bokeh Application\",\"version\":\"2.4.2\"}};\n  const render_items = [{\"docid\":\"63797384-4cb7-49d3-95bb-7035cc245c5e\",\"root_ids\":[\"8883\"],\"roots\":{\"8883\":\"000c7db5-5cc2-48b9-bcca-8fed60624189\"}}];\n  root.Bokeh.embed.embed_items_notebook(docs_json, render_items);\n\n  }\n  if (root.Bokeh !== undefined) {\n    embed_document(root);\n  } else {\n    let attempts = 0;\n    const timer = setInterval(function(root) {\n      if (root.Bokeh !== undefined) {\n        clearInterval(timer);\n        embed_document(root);\n      } else {\n        attempts++;\n        if (attempts > 100) {\n          clearInterval(timer);\n          console.log(\"Bokeh: ERROR: Unable to run BokehJS code because BokehJS library is missing\");\n        }\n      }\n    }, 10, root)\n  }\n})(window);",
      "application/vnd.bokehjs_exec.v0+json": ""
     },
     "metadata": {
      "application/vnd.bokehjs_exec.v0+json": {
       "id": "8883"
      }
     },
     "output_type": "display_data"
    }
   ],
   "source": [
    "import heapq\n",
    "from IPython.display import clear_output\n",
    "# generate maze from other notebook\n",
    "# Run 008-maze-generation-kruskal notebook\n",
    "%run 008-maze-generation-kruskal.ipynb"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Algorithm"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 106,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Node:\n",
    "  def __init__(self, g: int = 0, h: int = 0, f: int = float(\"inf\"), pos: list = []):\n",
    "    \"\"\"Representation of a cell, vertex or node\n",
    "\n",
    "    Args:\n",
    "        g (int): distance between node and source node\n",
    "        h (int): heuristic, estimated distance between node and destination\n",
    "        f (int): total cost of the node\n",
    "        pos (list): [y,x] position of the node in the maze\n",
    "    \"\"\"\n",
    "    self.g = g\n",
    "    self.h = h\n",
    "    self.f = f\n",
    "    self.pos = pos\n",
    "\n",
    "  def __eq__(self, other):\n",
    "    return (self.f == other.f)\n",
    "\n",
    "  def __ne__(self, other):\n",
    "    return not (self.f == other.f)\n",
    "\n",
    "  def __lt__(self, other):\n",
    "    return (self.f < other.f)\n",
    "\n",
    "  def __gt__(self, other):\n",
    "    return (self.f > other.f)\n",
    "\n",
    "  def __le__(self, other):\n",
    "    return (self.f <= other.f)\n",
    "\n",
    "  def __ge__(self, other):\n",
    "    return (self.f >= other.f)\n",
    "\n",
    "  def printNode(self, prefix: str = \"\", postfix: str = \"\"):\n",
    "    print(\"{}Node: [{},{}], g:{}, h:{} f:{}{}\".format(prefix, self.pos[0], self.pos[1], self.g, self.h, self.f, postfix))\n",
    "\n",
    "\n",
    "def aStar(maze, src: list, dest: list):\n",
    "  \"\"\"Function to find the shortest path between source cell and destination cell with help of the A* algorithm\n",
    "\n",
    "  Args:\n",
    "      maze (list): 2 dimensional maze matrix\n",
    "      src (list): y-x coordinates of source cell\n",
    "      dest (list): y-x coordinates of destination cell\n",
    "\n",
    "  Returns:\n",
    "      list: [distance of the solution, list of the maze solution path]\n",
    "  \"\"\"\n",
    "  # These arrays show the 4 possible movement from a cell\n",
    "  _rowNum = [-1, 0, 0, 1]\n",
    "  _colNum = [0, -1, 1, 0]\n",
    "\n",
    "  # store infinity\n",
    "  _infinity = float(\"inf\")\n",
    "\n",
    "  # Checking if source and destination cell have value 1\n",
    "  if maze[src[0]][src[1]]!=1 or maze[dest[0]][dest[1]]!=1:\n",
    "    return None\n",
    "\n",
    "  # get maze size\n",
    "  _maze_rows = len(maze)\n",
    "  _maze_cols = len(maze[0])\n",
    "\n",
    "  # list of visited cells\n",
    "  _visited = [[False for i in range(_maze_cols)]\n",
    "             for j in range(_maze_rows)]\n",
    "  # list of distances fill with infinity\n",
    "  _f = [[_infinity for i in range(_maze_cols)]\n",
    "               for j in range(_maze_rows)]\n",
    "\n",
    "  # Create a queue\n",
    "  _unvisited = []\n",
    "\n",
    "  # create source node and mark as visited\n",
    "  _visited[src[0]][src[1]] = True\n",
    "  # manhatten distance to dst\n",
    "  _h = abs(dest[0] - src[0]) + abs(dest[1] - src[1])\n",
    "  # zero distance to source\n",
    "  _g = 0\n",
    "  # zero cost\n",
    "  _f[src[0]][src[1]] = 0\n",
    "  _node = Node(_g, _h, _f[src[0]][src[1]], src)\n",
    "\n",
    "  # enqueue src node\n",
    "  heapq.heappush(_unvisited, _node)\n",
    "  #_node.printNode(\"Push \")\n",
    "\n",
    "  # Perform A* starting from source cell\n",
    "  while _unvisited:\n",
    "    # shortest unexplored path\n",
    "    _cur_node = heapq.heappop(_unvisited)\n",
    "    #_cur_node.printNode(\"Pop \")\n",
    "\n",
    "    # add node to visited cells\n",
    "    _visited[_cur_node.pos[0]][_cur_node.pos[1]] = True\n",
    "\n",
    "    # If we have reached the destination cell, return the final distance\n",
    "    # return f distances in order to backtrack\n",
    "    if _cur_node.pos[0] == dest[0] and _cur_node.pos[1] == dest[1]:\n",
    "      return _f\n",
    "\n",
    "    # Extend path to all adjacent cells\n",
    "    for i in range(4):\n",
    "      _adj_node_pos = (_cur_node.pos[0] + _rowNum[i], _cur_node.pos[1] + _colNum[i])\n",
    "\n",
    "      # if not visited yet and within maze and a cell (1))\n",
    "      if (not _visited[_adj_node_pos[0]][_adj_node_pos[1]] and\n",
    "         ((_adj_node_pos[0] >= 0) and (_adj_node_pos[0] < _maze_rows) and (_adj_node_pos[1] >= 0) and (_adj_node_pos[1] < _maze_cols)) and\n",
    "         (maze[_adj_node_pos[0]][_adj_node_pos[1]] == 1)):\n",
    "        # calculate f, g and h values\n",
    "        # calculate manhattan distance from cur_node to adj_node. Always one because it is a maze\n",
    "        _d = abs(_cur_node.pos[0] - _adj_node_pos[0]) + abs(_cur_node.pos[1] - _adj_node_pos[1])\n",
    "\n",
    "        # new distance\n",
    "        _g = _cur_node.g + _d\n",
    "        # manhattan distance from adj_node to dest\n",
    "        _h = abs(dest[0] - _adj_node_pos[0]) + abs(dest[1] - _adj_node_pos[1])\n",
    "        _f[_adj_node_pos[0]][_adj_node_pos[1]] = _g+_h\n",
    "        _adj_node = Node(_g, _h, _f[_adj_node_pos[0]][_adj_node_pos[1]], _adj_node_pos)\n",
    "\n",
    "        # see if node pos already existing in list (this should be improved)\n",
    "        _in_list = False\n",
    "        for node in _unvisited:\n",
    "          if node.pos == _adj_node.pos:\n",
    "            if _adj_node.g > node.g:\n",
    "              _in_list = True\n",
    "        if not _in_list:\n",
    "          heapq.heappush(_unvisited, _adj_node)\n",
    "          #_adj_node.printNode(\"Push \")\n",
    "\n",
    "  # Return none if destination cannot be reached\n",
    "  return None\n",
    "\n",
    "\n",
    "def getSolution(src: list, dst: list,  f: list):\n",
    "  \"\"\"Get shortest solution path\n",
    "\n",
    "  Args:\n",
    "      src (list): y,x coordinates of the startpoint\n",
    "      f (list): array of costs for each cell to arrive at the destination\n",
    "\n",
    "  Returns:\n",
    "      list: optimum solution path\n",
    "\n",
    "  \"\"\"\n",
    "  _solution_path = []\n",
    "  # These arrays show the 4 possible movement from a cell\n",
    "  _rowNum = [-1, 0, 0, 1]\n",
    "  _colNum = [0, -1, 1, 0]\n",
    "\n",
    "  # get maze size\n",
    "  _maze_rows = len(f)\n",
    "  _maze_cols = len(f[0])\n",
    "  _maze_size = _maze_cols*_maze_rows\n",
    "\n",
    "  _cur_pos = dst\n",
    "  _prev_pos = dst\n",
    "  # get maximum distance\n",
    "  _arrived = False\n",
    "  j = 0\n",
    "  while(not _arrived and j <= _maze_size):\n",
    "    _cost = float(\"inf\")\n",
    "    # for all 4 adjacent cells select the one with the smallest cost\n",
    "    for i in range(len(_rowNum)):\n",
    "      _next_pos = [_cur_pos[0] + _rowNum[i], _cur_pos[1] + _colNum[i]]\n",
    "      # check in on the board\n",
    "      if (_next_pos[0] >= 0) and (_next_pos[0] < _maze_rows) and (_next_pos[1] >= 0) and (_next_pos[1] < _maze_cols):\n",
    "        # Check if not the previous cell\n",
    "        if not (_prev_pos == _next_pos):\n",
    "          if f[_next_pos[0]][_next_pos[1]] < _cost:\n",
    "            _temp_pos = _next_pos\n",
    "            _cost = f[_next_pos[0]][_next_pos[1]]\n",
    "    if _cost != float(\"inf\"):\n",
    "      _solution_path.append(_temp_pos)\n",
    "      _prev_pos = _cur_pos\n",
    "      _cur_pos = _temp_pos\n",
    "\n",
    "    if _cur_pos == src:\n",
    "      return _solution_path\n",
    "    j += 1\n",
    "\n",
    "\n",
    "def drawSolution(maze: list, src: list, dst: list, solutionpath: list):\n",
    "  \"\"\"draw the start- and endpoint as well as the solution\n",
    "     0 = wall\n",
    "     1 = cell\n",
    "     2 = path\n",
    "     3 = start- endpoint\n",
    "\n",
    "  Args:\n",
    "      maze (list): 2 dimensional array of the maze\n",
    "      src (list): y,x coordinates of the startpoint\n",
    "      dst (list): y,x coordinates of the endpoint\n",
    "      solutionpath (list): array of points of the solution\n",
    "\n",
    "  Returns:\n",
    "      list: 2d array of the maze including solution\n",
    "\n",
    "  \"\"\"\n",
    "  maze_tmp = maze.copy()\n",
    "  # draw solution\n",
    "  for point in solutionpath:\n",
    "    maze_tmp[point[0]][point[1]] = 2\n",
    "  # draw start and endpoint\n",
    "  maze_tmp[src[0]][src[1]] = 3\n",
    "  maze_tmp[dst[0]][dst[1]] = 3\n",
    "  return maze_tmp\n",
    "\n",
    "def findEmptyCell(maze: list, corner: int):\n",
    "  \"\"\"find empty cell in cell starting at a corner\n",
    "\n",
    "  Args:\n",
    "      maze (list): maze to search through\n",
    "      corner (int): where to start (0=bottomleft, 1=topleft, 2=topright, 3=bottomright)\n",
    "\n",
    "  Returns:\n",
    "      list: row,col coordinates of the first empty cell\n",
    "  \"\"\"\n",
    "  maze_rows = len(maze)\n",
    "  maze_cols = len(maze[0])\n",
    "\n",
    "  if corner == 0:\n",
    "    range_rows = range(0, maze_rows-1, 1)\n",
    "    range_cols = range(0, maze_cols-1, 1)\n",
    "  elif corner == 1:\n",
    "    range_rows = range(0, maze_rows-1, 1)\n",
    "    range_cols = range(maze_cols-1, 0, -1)\n",
    "  elif corner == 2:\n",
    "    range_rows = range(maze_rows-1, 0, -1)\n",
    "    range_cols = range(maze_cols-1, 0, -1)\n",
    "  else:\n",
    "    range_rows = range(maze_rows-1, 0, -1)\n",
    "    range_cols = range(0, maze_cols-1, 1)\n",
    "\n",
    "  # find empty cell\n",
    "  for row in range_rows:\n",
    "    for col in range_cols:\n",
    "      if maze[row][col] == 1:  #  valid cell found\n",
    "        return  [row, col]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Test"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 107,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Maze Size           : 81x41\n",
      "Source Cell is      : [1, 10]\n",
      "Destination Cell is : [39, 76]\n",
      "Length of the Shortest Path is 144\n"
     ]
    }
   ],
   "source": [
    "no_solution = True\n",
    "while no_solution:\n",
    "\n",
    "  # generate maze with function from 008 notebook\n",
    "  size = {'x': 80,'y': 40}\n",
    "  maze = generateMaze(size['x'], size['y'])\n",
    "  maze = createBorder(maze)\n",
    "\n",
    "  # get maze size\n",
    "  maze_rows = len(maze)\n",
    "  maze_cols = len(maze[0])\n",
    "\n",
    "  # find empty cell in topleft corner\n",
    "  src = findEmptyCell(maze, 0)\n",
    "\n",
    "  # find empty cell in bottomright corner\n",
    "  dst = findEmptyCell(maze, 2)\n",
    "\n",
    "  print(\"Maze Size           : {}x{}\".format(maze_cols, maze_rows))\n",
    "  print(\"Source Cell is      : {}\".format(src))\n",
    "  print(\"Destination Cell is : {}\".format(dst))\n",
    "\n",
    "  # execute A* algorithm\n",
    "  f = aStar(maze, src, dst)\n",
    "\n",
    "  # draw points and solution\n",
    "  if f:\n",
    "    no_solution = False\n",
    "    path = getSolution(src, dst, f)\n",
    "    print(\"Length of the Shortest Path is\", len(path))\n",
    "    maze_solved = drawSolution(maze, src, dst, path)\n",
    "  else:\n",
    "    #no_solution = False##\n",
    "    print(\"No solution exists\")\n",
    "    clear_output(wait=True)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 108,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "\n",
       "    <div class=\"bk-root\">\n",
       "        <a href=\"https://bokeh.org\" target=\"_blank\" class=\"bk-logo bk-logo-small bk-logo-notebook\"></a>\n",
       "        <span id=\"9311\">Loading BokehJS ...</span>\n",
       "    </div>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "application/javascript": "\n(function(root) {\n  function now() {\n    return new Date();\n  }\n\n  const force = true;\n\n  if (typeof root._bokeh_onload_callbacks === \"undefined\" || force === true) {\n    root._bokeh_onload_callbacks = [];\n    root._bokeh_is_loading = undefined;\n  }\n\n  const JS_MIME_TYPE = 'application/javascript';\n  const HTML_MIME_TYPE = 'text/html';\n  const EXEC_MIME_TYPE = 'application/vnd.bokehjs_exec.v0+json';\n  const CLASS_NAME = 'output_bokeh rendered_html';\n\n  /**\n   * Render data to the DOM node\n   */\n  function render(props, node) {\n    const script = document.createElement(\"script\");\n    node.appendChild(script);\n  }\n\n  /**\n   * Handle when an output is cleared or removed\n   */\n  function handleClearOutput(event, handle) {\n    const cell = handle.cell;\n\n    const id = cell.output_area._bokeh_element_id;\n    const server_id = cell.output_area._bokeh_server_id;\n    // Clean up Bokeh references\n    if (id != null && id in Bokeh.index) {\n      Bokeh.index[id].model.document.clear();\n      delete Bokeh.index[id];\n    }\n\n    if (server_id !== undefined) {\n      // Clean up Bokeh references\n      const cmd_clean = \"from bokeh.io.state import curstate; print(curstate().uuid_to_server['\" + server_id + \"'].get_sessions()[0].document.roots[0]._id)\";\n      cell.notebook.kernel.execute(cmd_clean, {\n        iopub: {\n          output: function(msg) {\n            const id = msg.content.text.trim();\n            if (id in Bokeh.index) {\n              Bokeh.index[id].model.document.clear();\n              delete Bokeh.index[id];\n            }\n          }\n        }\n      });\n      // Destroy server and session\n      const cmd_destroy = \"import bokeh.io.notebook as ion; ion.destroy_server('\" + server_id + \"')\";\n      cell.notebook.kernel.execute(cmd_destroy);\n    }\n  }\n\n  /**\n   * Handle when a new output is added\n   */\n  function handleAddOutput(event, handle) {\n    const output_area = handle.output_area;\n    const output = handle.output;\n\n    // limit handleAddOutput to display_data with EXEC_MIME_TYPE content only\n    if ((output.output_type != \"display_data\") || (!Object.prototype.hasOwnProperty.call(output.data, EXEC_MIME_TYPE))) {\n      return\n    }\n\n    const toinsert = output_area.element.find(\".\" + CLASS_NAME.split(' ')[0]);\n\n    if (output.metadata[EXEC_MIME_TYPE][\"id\"] !== undefined) {\n      toinsert[toinsert.length - 1].firstChild.textContent = output.data[JS_MIME_TYPE];\n      // store reference to embed id on output_area\n      output_area._bokeh_element_id = output.metadata[EXEC_MIME_TYPE][\"id\"];\n    }\n    if (output.metadata[EXEC_MIME_TYPE][\"server_id\"] !== undefined) {\n      const bk_div = document.createElement(\"div\");\n      bk_div.innerHTML = output.data[HTML_MIME_TYPE];\n      const script_attrs = bk_div.children[0].attributes;\n      for (let i = 0; i < script_attrs.length; i++) {\n        toinsert[toinsert.length - 1].firstChild.setAttribute(script_attrs[i].name, script_attrs[i].value);\n        toinsert[toinsert.length - 1].firstChild.textContent = bk_div.children[0].textContent\n      }\n      // store reference to server id on output_area\n      output_area._bokeh_server_id = output.metadata[EXEC_MIME_TYPE][\"server_id\"];\n    }\n  }\n\n  function register_renderer(events, OutputArea) {\n\n    function append_mime(data, metadata, element) {\n      // create a DOM node to render to\n      const toinsert = this.create_output_subarea(\n        metadata,\n        CLASS_NAME,\n        EXEC_MIME_TYPE\n      );\n      this.keyboard_manager.register_events(toinsert);\n      // Render to node\n      const props = {data: data, metadata: metadata[EXEC_MIME_TYPE]};\n      render(props, toinsert[toinsert.length - 1]);\n      element.append(toinsert);\n      return toinsert\n    }\n\n    /* Handle when an output is cleared or removed */\n    events.on('clear_output.CodeCell', handleClearOutput);\n    events.on('delete.Cell', handleClearOutput);\n\n    /* Handle when a new output is added */\n    events.on('output_added.OutputArea', handleAddOutput);\n\n    /**\n     * Register the mime type and append_mime function with output_area\n     */\n    OutputArea.prototype.register_mime_type(EXEC_MIME_TYPE, append_mime, {\n      /* Is output safe? */\n      safe: true,\n      /* Index of renderer in `output_area.display_order` */\n      index: 0\n    });\n  }\n\n  // register the mime type if in Jupyter Notebook environment and previously unregistered\n  if (root.Jupyter !== undefined) {\n    const events = require('base/js/events');\n    const OutputArea = require('notebook/js/outputarea').OutputArea;\n\n    if (OutputArea.prototype.mime_types().indexOf(EXEC_MIME_TYPE) == -1) {\n      register_renderer(events, OutputArea);\n    }\n  }\n\n  \n  if (typeof (root._bokeh_timeout) === \"undefined\" || force === true) {\n    root._bokeh_timeout = Date.now() + 5000;\n    root._bokeh_failed_load = false;\n  }\n\n  const NB_LOAD_WARNING = {'data': {'text/html':\n     \"<div style='background-color: #fdd'>\\n\"+\n     \"<p>\\n\"+\n     \"BokehJS does not appear to have successfully loaded. If loading BokehJS from CDN, this \\n\"+\n     \"may be due to a slow or bad network connection. Possible fixes:\\n\"+\n     \"</p>\\n\"+\n     \"<ul>\\n\"+\n     \"<li>re-rerun `output_notebook()` to attempt to load from CDN again, or</li>\\n\"+\n     \"<li>use INLINE resources instead, as so:</li>\\n\"+\n     \"</ul>\\n\"+\n     \"<code>\\n\"+\n     \"from bokeh.resources import INLINE\\n\"+\n     \"output_notebook(resources=INLINE)\\n\"+\n     \"</code>\\n\"+\n     \"</div>\"}};\n\n  function display_loaded() {\n    const el = document.getElementById(\"9311\");\n    if (el != null) {\n      el.textContent = \"BokehJS is loading...\";\n    }\n    if (root.Bokeh !== undefined) {\n      if (el != null) {\n        el.textContent = \"BokehJS \" + root.Bokeh.version + \" successfully loaded.\";\n      }\n    } else if (Date.now() < root._bokeh_timeout) {\n      setTimeout(display_loaded, 100)\n    }\n  }\n\n\n  function run_callbacks() {\n    try {\n      root._bokeh_onload_callbacks.forEach(function(callback) {\n        if (callback != null)\n          callback();\n      });\n    } finally {\n      delete root._bokeh_onload_callbacks\n    }\n    console.debug(\"Bokeh: all callbacks have finished\");\n  }\n\n  function load_libs(css_urls, js_urls, callback) {\n    if (css_urls == null) css_urls = [];\n    if (js_urls == null) js_urls = [];\n\n    root._bokeh_onload_callbacks.push(callback);\n    if (root._bokeh_is_loading > 0) {\n      console.debug(\"Bokeh: BokehJS is being loaded, scheduling callback at\", now());\n      return null;\n    }\n    if (js_urls == null || js_urls.length === 0) {\n      run_callbacks();\n      return null;\n    }\n    console.debug(\"Bokeh: BokehJS not loaded, scheduling load and callback at\", now());\n    root._bokeh_is_loading = css_urls.length + js_urls.length;\n\n    function on_load() {\n      root._bokeh_is_loading--;\n      if (root._bokeh_is_loading === 0) {\n        console.debug(\"Bokeh: all BokehJS libraries/stylesheets loaded\");\n        run_callbacks()\n      }\n    }\n\n    function on_error(url) {\n      console.error(\"failed to load \" + url);\n    }\n\n    for (let i = 0; i < css_urls.length; i++) {\n      const url = css_urls[i];\n      const element = document.createElement(\"link\");\n      element.onload = on_load;\n      element.onerror = on_error.bind(null, url);\n      element.rel = \"stylesheet\";\n      element.type = \"text/css\";\n      element.href = url;\n      console.debug(\"Bokeh: injecting link tag for BokehJS stylesheet: \", url);\n      document.body.appendChild(element);\n    }\n\n    for (let i = 0; i < js_urls.length; i++) {\n      const url = js_urls[i];\n      const element = document.createElement('script');\n      element.onload = on_load;\n      element.onerror = on_error.bind(null, url);\n      element.async = false;\n      element.src = url;\n      console.debug(\"Bokeh: injecting script tag for BokehJS library: \", url);\n      document.head.appendChild(element);\n    }\n  };\n\n  function inject_raw_css(css) {\n    const element = document.createElement(\"style\");\n    element.appendChild(document.createTextNode(css));\n    document.body.appendChild(element);\n  }\n\n  \n  const js_urls = [\"https://cdn.bokeh.org/bokeh/release/bokeh-2.4.2.min.js\", \"https://cdn.bokeh.org/bokeh/release/bokeh-gl-2.4.2.min.js\", \"https://cdn.bokeh.org/bokeh/release/bokeh-widgets-2.4.2.min.js\", \"https://cdn.bokeh.org/bokeh/release/bokeh-tables-2.4.2.min.js\", \"https://cdn.bokeh.org/bokeh/release/bokeh-mathjax-2.4.2.min.js\"];\n  const css_urls = [];\n  \n\n  const inline_js = [\n    function(Bokeh) {\n      Bokeh.set_log_level(\"info\");\n    },\n    function(Bokeh) {\n    \n    \n    }\n  ];\n\n  function run_inline_js() {\n    \n    if (root.Bokeh !== undefined || force === true) {\n      \n    for (let i = 0; i < inline_js.length; i++) {\n      inline_js[i].call(root, root.Bokeh);\n    }\n    if (force === true) {\n        display_loaded();\n      }} else if (Date.now() < root._bokeh_timeout) {\n      setTimeout(run_inline_js, 100);\n    } else if (!root._bokeh_failed_load) {\n      console.log(\"Bokeh: BokehJS failed to load within specified timeout.\");\n      root._bokeh_failed_load = true;\n    } else if (force !== true) {\n      const cell = $(document.getElementById(\"9311\")).parents('.cell').data().cell;\n      cell.output_area.append_execute_result(NB_LOAD_WARNING)\n    }\n\n  }\n\n  if (root._bokeh_is_loading === 0) {\n    console.debug(\"Bokeh: BokehJS loaded, going straight to plotting\");\n    run_inline_js();\n  } else {\n    load_libs(css_urls, js_urls, function() {\n      console.debug(\"Bokeh: BokehJS plotting callback run at\", now());\n      run_inline_js();\n    });\n  }\n}(window));",
      "application/vnd.bokehjs_load.v0+json": ""
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "  <div class=\"bk-root\" id=\"dc185b44-8eec-48ea-a8a2-b6d40903e2c6\" data-root-id=\"9312\"></div>\n"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "application/javascript": "(function(root) {\n  function embed_document(root) {\n    \n  const docs_json = {\"d59d9534-6e9b-4ca5-a95c-2a69be12e1f7\":{\"defs\":[],\"roots\":{\"references\":[{\"attributes\":{\"below\":[{\"id\":\"9321\"}],\"center\":[{\"id\":\"9324\"},{\"id\":\"9328\"}],\"height\":410,\"left\":[{\"id\":\"9325\"}],\"outline_line_color\":\"#ffffff\",\"renderers\":[{\"id\":\"9353\"}],\"title\":{\"id\":\"9685\"},\"toolbar\":{\"id\":\"9336\"},\"width\":810,\"x_range\":{\"id\":\"9313\"},\"x_scale\":{\"id\":\"9317\"},\"y_range\":{\"id\":\"9315\"},\"y_scale\":{\"id\":\"9319\"}},\"id\":\"9312\",\"subtype\":\"Figure\",\"type\":\"Plot\"},{\"attributes\":{},\"id\":\"9332\",\"type\":\"SaveTool\"},{\"attributes\":{\"color_mapper\":{\"id\":\"9348\"},\"dh\":{\"value\":1},\"dw\":{\"value\":1},\"global_alpha\":{\"value\":0.1},\"image\":{\"field\":\"image\"},\"x\":{\"value\":0},\"y\":{\"value\":0}},\"id\":\"9347\",\"type\":\"Image\"},{\"attributes\":{\"palette\":[\"#228b22\",\"#ffffff\",\"#add8e6\",\"#ff7f7f\"]},\"id\":\"9348\",\"type\":\"LinearColorMapper\"},{\"attributes\":{},\"id\":\"9689\",\"type\":\"AllLabels\"},{\"attributes\":{\"bottom_units\":\"screen\",\"coordinates\":null,\"fill_alpha\":0.5,\"fill_color\":\"lightgrey\",\"group\":null,\"left_units\":\"screen\",\"level\":\"overlay\",\"line_alpha\":1.0,\"line_color\":\"black\",\"line_dash\":[4,4],\"line_width\":2,\"right_units\":\"screen\",\"syncable\":false,\"top_units\":\"screen\"},\"id\":\"9335\",\"type\":\"BoxAnnotation\"},{\"attributes\":{},\"id\":\"9692\",\"type\":\"AllLabels\"},{\"attributes\":{\"tools\":[{\"id\":\"9329\"},{\"id\":\"9330\"},{\"id\":\"9331\"},{\"id\":\"9332\"},{\"id\":\"9333\"},{\"id\":\"9334\"}]},\"id\":\"9336\",\"type\":\"Toolbar\"},{\"attributes\":{\"axis\":{\"id\":\"9325\"},\"coordinates\":null,\"dimension\":1,\"group\":null,\"ticker\":null},\"id\":\"9328\",\"type\":\"Grid\"},{\"attributes\":{},\"id\":\"9334\",\"type\":\"HelpTool\"},{\"attributes\":{},\"id\":\"9693\",\"type\":\"UnionRenderers\"},{\"attributes\":{\"coordinates\":null,\"formatter\":{\"id\":\"9688\"},\"group\":null,\"major_label_policy\":{\"id\":\"9689\"},\"ticker\":{\"id\":\"9326\"},\"visible\":false},\"id\":\"9325\",\"type\":\"LinearAxis\"},{\"attributes\":{},\"id\":\"9329\",\"type\":\"PanTool\"},{\"attributes\":{},\"id\":\"9688\",\"type\":\"BasicTickFormatter\"},{\"attributes\":{},\"id\":\"9330\",\"type\":\"WheelZoomTool\"},{\"attributes\":{},\"id\":\"9691\",\"type\":\"BasicTickFormatter\"},{\"attributes\":{},\"id\":\"9322\",\"type\":\"BasicTicker\"},{\"attributes\":{\"data\":{\"image\":[[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0,0,0,3,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,0],[0,1,1,0,1,0,1,1,2,2,2,0,2,2,2,2,2,2,2,2,2,2,2,2,2,0,1,0,1,0,1,1,1,0,1,1,1,0,2,2,2,2,2,2,2,0,1,1,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0],[0,0,1,0,1,0,1,0,2,0,0,0,2,0,1,0,0,0,0,0,0,0,0,0,2,0,1,0,1,0,0,0,1,0,1,0,0,0,2,0,0,0,0,0,2,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,0],[0,0,1,0,1,1,1,0,2,1,1,0,2,0,1,1,1,1,1,1,1,0,1,1,2,0,1,1,1,1,1,1,1,1,1,0,1,1,2,1,1,0,1,0,2,2,2,2,2,0,1,1,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,0,0],[0,0,1,0,1,0,0,0,2,0,0,0,2,0,0,0,0,0,1,0,0,0,1,0,2,0,0,0,0,0,1,0,0,0,0,0,0,0,2,0,0,0,1,0,1,0,1,0,2,0,0,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,0],[0,0,1,1,1,0,1,0,2,2,2,2,2,0,1,1,1,1,1,1,1,0,1,0,2,2,2,2,2,2,2,0,1,0,1,1,1,0,2,0,1,1,1,1,1,0,1,0,2,0,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,0,1,0,0],[0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,2,0,1,0,1,0,0,0,2,0,0,0,1,0,0,0,0,0,2,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,0],[0,1,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,0,1,1,1,1,1,1,1,1,2,0,2,2,2,2,2,2,2,0,1,1,1,1,1,0,1,1,2,0,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,0,1,0,1,0,1,0,1,1,1,0,1,1,0],[0,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,2,0,2,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,2,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,1,0,0],[0,1,1,1,1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,2,2,2,0,1,1,1,0,1,0,1,1,1,0,1,1,1,0,2,1,1,0,1,0,1,0,1,0,1,1,1,1,1,0,1,0,2,2,2,0,1,1,1,1,1,0,1,1,1,1,0],[0,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,1,0,0,0,2,0,0,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,2,0,2,0,0,0,1,0,0,0,0,0,0,0,0],[0,1,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,1,1,0,1,0,2,2,2,0,1,0,1,1,1,0,1,0,1,0,1,0,2,2,2,0,2,0,1,1,1,0,1,1,1,1,1,0,0],[0,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,2,0,1,0,1,0,0,0,1,0,1,0,1,0,2,0,0,0,2,0,0,0,1,0,0,0,1,0,0,0,0],[0,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,0,1,1,2,1,1,1,1,1,1,1,1,1,1,1,0],[0,0,0,0,0,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,2,0,0,0,0,0,0,0,1,0,0,0,0],[0,1,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,0,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,0,2,2,2,1,1,0,1,1,1,1,1,1,0],[0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,0,0,1,0,0,0,0,0,2,0,0,0,0,0,1,0,0,0,0],[0,0,1,0,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,0,1,1,1,0,1,0,2,1,1,1,1,0,1,1,1,0,0],[0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,2,0,0,0,0,0,0,0,1,0,0],[0,1,1,0,1,0,1,0,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,0,2,2,2,0,1,0,1,0,1,0,0],[0,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,2,0,1,0,1,0,0,0,0],[0,1,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,1,2,0,1,1,1,1,1,1,0],[0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,0,0,1,0,0],[0,1,1,0,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,0,1,1,1,1,2,2,2,0,1,1,1,0,1,1,0],[0,0,1,0,1,0,1,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,0,0,2,0,0,0,0,0,1,0,0,0,0],[0,0,1,1,1,0,1,0,1,0,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,1,2,2,2,1,1,0,1,1,1,1,0],[0,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,0,0,1,0,0,0,0,0,2,0,0,0,1,0,0,0,0],[0,1,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,1,2,1,1,0,1,1,1,1,0],[0,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,2,0,1,0,1,0,0,0,0],[0,0,1,0,1,0,1,0,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,1,1,0,2,0,1,1,1,1,1,0,0],[0,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0],[0,1,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,0,1,1,1,0,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,0,2,1,1,1,1,1,1,0,0],[0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,2,0,1,0,0,0,0,0,0],[0,1,1,1,1,1,1,1,1,1,1,0,1,0,1,1,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,0,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,0,2,0,1,0,1,1,1,0,0],[0,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,1,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,2,0,0,0,0,0,1,0,0],[0,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,0,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,0,1,0,1,1,1,1,1,0,1,0,1,1,1,0,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,2,2,2,0,1,0,1,1,0],[0,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,2,0,1,0,0,0,0],[0,1,1,1,1,0,1,1,1,1,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,0,1,1,1,1,1,0,2,2,2,1,1,1,0],[0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,1,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,1,0,0,0,3,0,0,0,0],[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]]]},\"selected\":{\"id\":\"9694\"},\"selection_policy\":{\"id\":\"9693\"}},\"id\":\"9343\",\"type\":\"ColumnDataSource\"},{\"attributes\":{\"coordinates\":null,\"data_source\":{\"id\":\"9343\"},\"glyph\":{\"id\":\"9344\"},\"group\":null,\"hover_glyph\":null,\"muted_glyph\":{\"id\":\"9350\"},\"nonselection_glyph\":{\"id\":\"9347\"},\"view\":{\"id\":\"9354\"}},\"id\":\"9353\",\"type\":\"GlyphRenderer\"},{\"attributes\":{\"coordinates\":null,\"group\":null},\"id\":\"9685\",\"type\":\"Title\"},{\"attributes\":{\"overlay\":{\"id\":\"9335\"}},\"id\":\"9331\",\"type\":\"BoxZoomTool\"},{\"attributes\":{},\"id\":\"9694\",\"type\":\"Selection\"},{\"attributes\":{},\"id\":\"9333\",\"type\":\"ResetTool\"},{\"attributes\":{\"axis\":{\"id\":\"9321\"},\"coordinates\":null,\"group\":null,\"ticker\":null},\"id\":\"9324\",\"type\":\"Grid\"},{\"attributes\":{},\"id\":\"9315\",\"type\":\"Range1d\"},{\"attributes\":{\"coordinates\":null,\"formatter\":{\"id\":\"9691\"},\"group\":null,\"major_label_policy\":{\"id\":\"9692\"},\"ticker\":{\"id\":\"9322\"},\"visible\":false},\"id\":\"9321\",\"type\":\"LinearAxis\"},{\"attributes\":{},\"id\":\"9317\",\"type\":\"LinearScale\"},{\"attributes\":{},\"id\":\"9319\",\"type\":\"LinearScale\"},{\"attributes\":{\"palette\":[\"#228b22\",\"#ffffff\",\"#add8e6\",\"#ff7f7f\"]},\"id\":\"9345\",\"type\":\"LinearColorMapper\"},{\"attributes\":{\"color_mapper\":{\"id\":\"9351\"},\"dh\":{\"value\":1},\"dw\":{\"value\":1},\"global_alpha\":{\"value\":0.2},\"image\":{\"field\":\"image\"},\"x\":{\"value\":0},\"y\":{\"value\":0}},\"id\":\"9350\",\"type\":\"Image\"},{\"attributes\":{\"color_mapper\":{\"id\":\"9345\"},\"dh\":{\"value\":1},\"dw\":{\"value\":1},\"image\":{\"field\":\"image\"},\"x\":{\"value\":0},\"y\":{\"value\":0}},\"id\":\"9344\",\"type\":\"Image\"},{\"attributes\":{\"source\":{\"id\":\"9343\"}},\"id\":\"9354\",\"type\":\"CDSView\"},{\"attributes\":{\"palette\":[\"#228b22\",\"#ffffff\",\"#add8e6\",\"#ff7f7f\"]},\"id\":\"9351\",\"type\":\"LinearColorMapper\"},{\"attributes\":{},\"id\":\"9313\",\"type\":\"Range1d\"},{\"attributes\":{},\"id\":\"9326\",\"type\":\"BasicTicker\"}],\"root_ids\":[\"9312\"]},\"title\":\"Bokeh Application\",\"version\":\"2.4.2\"}};\n  const render_items = [{\"docid\":\"d59d9534-6e9b-4ca5-a95c-2a69be12e1f7\",\"root_ids\":[\"9312\"],\"roots\":{\"9312\":\"dc185b44-8eec-48ea-a8a2-b6d40903e2c6\"}}];\n  root.Bokeh.embed.embed_items_notebook(docs_json, render_items);\n\n  }\n  if (root.Bokeh !== undefined) {\n    embed_document(root);\n  } else {\n    let attempts = 0;\n    const timer = setInterval(function(root) {\n      if (root.Bokeh !== undefined) {\n        clearInterval(timer);\n        embed_document(root);\n      } else {\n        attempts++;\n        if (attempts > 100) {\n          clearInterval(timer);\n          console.log(\"Bokeh: ERROR: Unable to run BokehJS code because BokehJS library is missing\");\n        }\n      }\n    }, 10, root)\n  }\n})(window);",
      "application/vnd.bokehjs_exec.v0+json": ""
     },
     "metadata": {
      "application/vnd.bokehjs_exec.v0+json": {
       "id": "9312"
      }
     },
     "output_type": "display_data"
    }
   ],
   "source": [
    "output_notebook()\n",
    "\n",
    "plot = figure(x_range=(0, 1), y_range=(0, 1),\n",
    "              plot_height=maze_rows*10, plot_width=maze_cols*10)\n",
    "plot.axis.visible = False\n",
    "plot.outline_line_color = '#ffffff'\n",
    "\n",
    "plot.image([maze_solved], x=0, y=0, dw=1, dh=1, palette=['#228b22', '#ffffff', '#add8e6', '#ff7f7f'])\n",
    "\n",
    "show(plot)"
   ]
  }
 ],
 "metadata": {
  "interpreter": {
   "hash": "579658931a9ee5a566d859011b110d45b1307b50642b87f1e088320bdb784c2c"
  },
  "kernelspec": {
   "display_name": "Python 3.8.11 64-bit ('daysofalgo': conda)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.11"
  },
  "orig_nbformat": 4
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
